
import json
import datetime

import astropy.units as u
from astropy.coordinates import SkyCoord, ICRS, FK4
from . import def_name

import re

import logging
log = logging.getLogger(__name__)


TIME_FMT = "%Y-%m-%dT%H:%M:%S.%fZ"


def append_date_subquery(curr_query, params):
    """ Append a date search to the query string.

    :param curr_query: The current query string we're going to append to and return.
    :param params: The parameter dictionary which may or may not contain populated date fields.
    :return: If there are no values, the original string, otherwise the original string with the new query parameters
    appended in solr query form:

    - If no max date is provided, the search is assumed to be from min to ALL after
    - If no min date is provided, the search is assumed to be from ALL before to max
    - If both min and max are given, the search is a window from min to max

    Min dates are assumed from the start of the day and max dates from the end of the day
    """
    q = [curr_query]
    try:
        if params.mindate and params.maxdate is None:
            # All observations where obs_stop is greater than desired start
            min_date = datetime.datetime.strptime(''.join([params.mindate, 'T00:00:00.000Z']), TIME_FMT)
            if isinstance(min_date, datetime.datetime):
                q.append(f' AND obs_stop:[{min_date.strftime(TIME_FMT)} TO *]')
        elif params.maxdate and params.mindate is None:
            # All observations where obs_start is less than desired end
            max_date = datetime.datetime.strptime(''.join([params.maxdate, 'T23:59:59.999Z']), TIME_FMT)
            if isinstance(max_date, datetime.datetime):
                q.append(f' AND obs_start:[* TO {max_date.strftime(TIME_FMT)}]')
        elif params.mindate and params.maxdate:
            # All observation:
            #  - completely within the desired window   (obs_start > start and obs_stop < end)
            #  - straddling the windows bounds          (obs_start < start and obs_stop > end)
            #  - stradding the desired start            (obs_start < start and obs_stop < end and obs_stop > start)
            #  - stradding the desired end              (obs_start > start and obs_stop > end and obs_start < end)
            min_date = datetime.datetime.strptime(''.join([params.mindate, 'T00:00:00.000Z']), TIME_FMT)
            max_date = datetime.datetime.strptime(''.join([params.maxdate, 'T23:59:59.999Z']), TIME_FMT)
            if isinstance(min_date, datetime.datetime) and isinstance(max_date, datetime.datetime):
                q.append(f' AND (')
                q.append(f'(obs_start:[{min_date.strftime(TIME_FMT)} TO *] AND '
                         f'obs_stop:[* TO {max_date.strftime(TIME_FMT)}])')
                q.append(' OR ')
                q.append(f'(obs_start:[* TO {min_date.strftime(TIME_FMT)}] AND '
                         f'obs_stop:[{max_date.strftime(TIME_FMT)} TO *])')
                q.append(' OR ')
                q.append(f'(obs_start:[* TO {min_date.strftime(TIME_FMT)}] AND '
                         f'obs_stop:[* TO {max_date.strftime(TIME_FMT)}] AND '
                         f'obs_stop:[{min_date.strftime(TIME_FMT)} TO *])')
                q.append(' OR ')
                q.append(f'(obs_start:[{min_date.strftime(TIME_FMT)} TO *] AND '
                         f'obs_stop:[{max_date.strftime(TIME_FMT)} TO *] AND '
                         f'obs_start:[* TO {max_date.strftime(TIME_FMT)}]))')
    except Exception as x:
        ex = 'Exception caught in {0}: {1}'.format(def_name(), str(x))
        raise Exception(ex)
    return ''.join(q)


def get_eb_position_filter(params):
    """ Construct a dictionary with the parameters solr needs for a positional search, using the geofilt feature.

    :param params: The parameter dictionary which may or may not contain populated positional search parameters
    :return: if there are no values, an empty dictionary, otherwise a dictionary with the parameters for a solr
    positional search populated:
    - fq = {!geofilt sfield=target_pos_<COORDINATE_REFERENCE_FRAME>}
    - d = the radius of the search
    - p = the coordinate pair for the "latitude, longitude" of the search.
    """
    geofilt = {}

    if params.longitude is not None and params.latitude is not None:

        if not params.coordsys_refFrame:
            geofilt['fq'] = '{!geofilt sfield=target_pos_equatorial}'
        else:
            if params.coordsys_refFrame in ['equatorial', 'ecliptic', 'galactic']:
                geofilt['fq'] = '{!geofilt sfield=target_pos_' + params.coordsys_refFrame + '}'
            else:
                ex = 'Exception caught in {0}: {1}'.format(def_name(),
                                                           'unrecognized coordinate reference frame')
                raise Exception(ex)

        geofilt['d'] = params.radius if params.radius else 1.0

        geofilt['pt'] = '{0}, {1}'.format(params.latitude, params.longitude)

    return geofilt


def get_alternate_eb_position_filter(curr_query, params):
    """ Construct a subquery for a positional search against the subscan positions, using the geofilt feature.

    :param curr_query: The current query string we're going to append to and return
    :param params: The parameter dictionary which may or may not contain populated positional search parameters
    :return: if there are no position values in the search parameter list, the current query string 'unchanged':
        return ''.join([curr_query])

        Otherwise, construct a query against the subscans of execution blocks.  Here we are building a rather
        long solr query, but it's a result of the conceptualization of the beam pattern and Field Of View (FOV).
        From the legacy archive:
        "   Scalable Search Radius:
            A search radius on the sky whose radius is equal the the half-power
            angular radius of a VLA antenna primary beam. This radius is scalable
            with the inverse of the observing frequency. Archive data within
            this radius surrounding the observed sky position is considered
            usable.

            The search radii for the VLA are :

            freq. band  |  search rad (arc-mins)
            -----------------------------------
             4       |   350.0
             P       |    75.0
             L       |    15.0
             C       |     4.5
             X       |     2.7
             U       |     1.5
             K       |     1.0
             Q       |     0.5

            For all VLBA observations, the scalable search radius is
            set to one arc-min.
        "

        The idea here is that for any position (or point with radius, ie FOVin) the user selects, we only want to
        return observations that conceptually would include the point(s) of interest.  A P band observation will
        include a larger portion of the sky (75') than a Q band (0.5'), so the user's input point(s) would more
        likely to be found in observations in P band than in Q band. Of course, that depends on band allocations over
        the years, but if each had exactly equal time, Q band would have far less sky coverage than P band.  Here are
        a couple examples to explain.

        Imagine we allocated equal time for a hypothetical "75' wide, galactic disk - planar survey" in P band and Q
        band.  Since the Q band FOV is 1/150th the FOV of a P band, the P band could survey it in one pass,
        and then another 150^2-1 times before a single pass had been completed in Q band (neglecting all
        time-on-source requirements due to frequency dependence and band width - they take a snapshot of an area and
        capture their entire bandwidth in the same time for illustrative purposes). This plays a little fast and
        loose with coordinate frames and areas of coverage, but imagine a single P band observation is ~150 Q band
        observations wide and ~150 observations long.

        Another way to look at the FOV query is to imagine looking for any data on a small circular patch in the sky,
        perhaps that has no known sources within it, a 'dark' patch.  The likelihood of that point being within the
        FOV for any Q band observations is small, unless the point selected is within 0.5' of a given Q band
        observation.  Whereas, in P band, we likely have large swaths of the sky captured, so finding observations
        that might include a random patch of sky within the FOV is a little higher.

        Hopefully you can now further understand why the solr query takes on the form of a {!geofilt} query per
        instrument per band.  If the point the user enters is not within the radius of the observation, based on the
        observations FOV, we aren't interested in it.

        q = q + AND
                (
                    (instrument_name:(VLA OR EVLA) AND
                        (
                            (obs_band:4 AND _query_:"{!parent which=_docType_:exec_block}
                                                        {!geofilt
                                                            sfield=sbsc_target_<input>
                                                            pt=<Dec,RA-180>
                                                            d=<4-band FOV> + <user_input_radius_in_degrees | 0>
                                                        }")
                             OR
                            (obs_band:P AND _query_:"{!parent which=_docType_:exec_block}
                                                        {!geofilt
                                                            sfield=sbsc_target_<input>
                                                            pt=<Dec,RA-180>
                                                            d=<P-band FOV> + <user_input_radius_in_degrees | 0>
                                                        }")
                             OR
                            ... <L, S, C, X, KU, K, KA, Q FOV> ...
                             OR
                            (-obs_band:(4 OR P OR L OR S OR C OR X OR KU OR K OR KA OR Q)
                                        AND _query_:"{!parent which=_docType_:exec_block}
                                                        {!geofilt
                                                            sfield=sbsc_target_<input>
                                                            pt=<Dec,RA-180>
                                                            d=<1' FOV> + <user_input_radius_in_degrees | 0>
                                                        }")
                        )
                    )
                     OR
                    (instrument_name:VLBA AND _query_:"{!parent which=_docType_:exec_block}
                                                {!geofilt
                                                    sfield=sbsc_target_<input>
                                                    pt=<Dec,RA-180>
                                                    d=<1' FOV> + <user_input_radius_in_degrees | 0>
                                                }"
                    )
                     OR
                    (instrument_name:ALMA AND _query_:"{!parent which=_docType_:exec_block}
                                                {!geofilt
                                                    sfield=sbsc_target_<input>
                                                    pt=<Dec,RA-180>
                                                    d=<1/2' FOV> + <user_input_radius_in_degrees | 0>
                                                }"
                    )
                )

    """
    q = [curr_query]
    if params.longitude is not None and params.latitude is not None:
        if params.coordsys_refFrame in ['equatorial', 'ecliptic', 'galactic']:
            geofilt_frame = 'sbsc_target_pos_' + params.coordsys_refFrame
        else:
            ex = 'Exception caught in {0}: {1}'.format(def_name(),
                                                       'unrecognized coordinate reference frame')
            raise Exception(ex)

        user_d = params.radius if params.radius else 0
        long = params.longitude
        lat = params.latitude

        # Our list mirroring the legacy archive FOV table in the help (these can be tweaked, or maybe capo-ified?)
        # https://archive.nrao.edu/archive/nraodashelp.html#autoSrad
        # according to that source, K has a FOV of 1', so, we use that radius below for 1' defaults
        VLA_FOV_list = [
            {'band': '4', 'r': 5.8333},
            {'band': 'P', 'r': 1.25},
            {'band': 'L', 'r': 0.25},
            {'band': 'S', 'r': 0.1333},
            {'band': 'C', 'r': 0.075},
            {'band': 'X', 'r': 0.045},
            {'band': 'KU', 'r': 0.025},
            {'band': 'K', 'r': 0.01666},
            {'band': 'KA', 'r': 0.0125},
            {'band': 'Q', 'r': 0.008333}
        ]

        q.append(' AND (')
        # VLA/EVLA
        q.append('(instrument_name:(VLA OR EVLA) AND (')
        vla_band_query = []
        vla_bands = []
        # Step through our list of fixed FOV radii (4,P,L,S,C,X,KU,K,KA,Q)
        for band in VLA_FOV_list:
            temp = []
            vla_bands.append(band['band'])
            temp.append('(obs_band:{0} AND _query_:"{{!parent which=_docType_:exec_block}}'.format(band['band']))
            temp.append(
                '{{!geofilt sfield={0} pt={1},{2} d={3}}}")'.format(geofilt_frame, lat, long, band['r'] + user_d))
            vla_band_query.append(''.join(temp))
        # WE've got our list of child position queries by FOV, join them on ' OR '
        q.append(' OR '.join(vla_band_query))
        # And OR a default of 1' for all VLA/EVLA bands NOT (4|P|L|S|C|X|KU|K|KA|Q)
        q.append(' OR ')
        q.append('(-obs_band:(')
        q.append(' OR '.join(vla_bands))
        q.append(') AND _query_:"{!parent which=_docType_:exec_block}')
        q.append('{{!geofilt sfield={0} pt={1},{2} d={3}}}")'.format(geofilt_frame, lat, long, 0.01666 + user_d))
        q.append('))')  # End VLA/EVLA
        # Assume 1' for VLBA observations... for now
        q.append(' OR (instrument_name:VLBA AND _query_:"{!parent which=_docType_:exec_block}')
        q.append('{{!geofilt sfield={0} pt={1},{2} d={3}}}")'.format(geofilt_frame, lat, long, 0.01666 + user_d))
        # And 0.5' for ALMA, because the dish is smaller, so, without more input, we'll assume half an arc-min
        q.append(' OR (instrument_name:ALMA AND _query_:"{!parent which=_docType_:exec_block}')
        q.append('{{!geofilt sfield={0} pt={1},{2} d={3}}}")'.format(geofilt_frame, lat, long, 0.008333 + user_d))
        q.append(')')
    return ''.join(q)


def get_images_position_filter(params):
    """ Construct a dictionary with the parameters solr needs for a positional search, using the geofilt feature.

    :param params: The parameter dictionary which may or may not contain populated positional search parameters
    :return: if there are no values, an empty dictionary, otherwise a dictionary with the parameters for a solr
    positional search populated:
    - fq = {!geofilt sfield=target_pos_<COORDINATE_REFERENCE_FRAME>}
    - d = the radius of the search
    - p = the coordinate pair for the "latitude, longitude" of the search.
    """
    geofilt = {}

    if params.longitude is not None and params.latitude is not None:

        if not params.coordsys_refFrame:
            geofilt['fq'] = '{!geofilt sfield=target_pos_equatorial}'
        else:
            if params.coordsys_refFrame in ['equatorial', 'ecliptic', 'galactic']:
                geofilt['fq'] = '{!geofilt sfield=target_pos_' + params.coordsys_refFrame + '}'
            else:
                ex = 'Exception caught in {0}: {1}'.format(def_name(),
                                                           'unrecognized coordinate reference frame')
                raise Exception(ex)

        geofilt['d'] = params.radius if params.radius else 1.0

        geofilt['pt'] = '{0}, {1}'.format(params.latitude, params.longitude)

    return geofilt


def append_eb_frequency_subquery(curr_query, params):
    """ Append a frequency search to the query string.

    :param curr_query: The current query string we're going to append to and return
    :param params: The parameter dictionary which may or may not contain populated frequency fields
    :return: If there are no values, the original string, otherwise the original string with the new query parameters
    appended in solr query form:

    Because frequency information is stored in child documents, this adds a sub-query to the query string.  And, like
    the date search query, we allow for frequency windows, or limits.
    - If no max frequency is provided, the search is assumed to be from min to ALL after
    - If no min frequency is provided, the search is assumed to be from ALL before to max
    - If both min and max are given, the search is a window from min to max
    """
    q = [curr_query]
    try:
        if params.startfreq and params.endfreq is None:
            # All observation where max_frequency is greater than desired start
            q.append(' AND _query_:\"{!parent which=_docType_:exec_block}')
            q.append(f'max_frequency:[{params.startfreq:.9e} TO *]\"')
        elif params.endfreq and params.startfreq is None:
            # All observations where min_frequency is less than desired end
            q.append(' AND _query_:\"{!parent which=_docType_:exec_block}')
            q.append(f'min_frequency:[* TO {params.endfreq:.9e}]\"')
        elif params.startfreq and params.endfreq:
            # All observation:
            #  - completely within the desired window   (min_freq > start and max_freq < end)
            #  - straddling the windows bounds          (min_freq < start and max_freq > end)
            #  - stradding the desired start            (min_freq < start and max_freq < end and max_freq > start)
            #  - stradding the desired end              (min_freq > start and max_freq > end and min_freq < end)
            q.append(' AND _query_:\"{!parent which=_docType_:exec_block}')
            q.append(f'((min_frequency:[{params.startfreq:.9e} TO *] AND max_frequency:[* TO {params.endfreq:.9e}])')
            q.append(' OR ')
            q.append(f'(min_frequency:[* TO {params.startfreq:.9e}] AND max_frequency:[{params.endfreq:.9e} TO *])')
            q.append(' OR ')
            q.append(f'(min_frequency:[* TO {params.startfreq:.9e}] AND max_frequency:[* TO {params.endfreq:.9e}]'
                     f' AND max_frequency:[{params.startfreq:.9e} TO *])')
            q.append(' OR ')
            q.append(f'(min_frequency:[{params.startfreq:.9e} TO *] AND max_frequency:[{params.endfreq:.9e} TO *]'
                     f' AND min_frequency:[* TO {params.endfreq:.9e}]))\"')
    except Exception as x:
        ex = 'Exception caught in {0}: {1}'.format(def_name(), str(x))
        raise Exception(ex)
    return ''.join(q)


def append_images_frequency_subquery(curr_query, params):
    """ Append a frequency search to the query string.

    :param curr_query: The current query string we're going to append to and return
    :param params: The parameter dictionary which may or may not contain populated frequency fields
    :return: If there are no values, the original string, otherwise the original string with the new query parameters
    appended in solr query form:

    - If no max frequency is provided, the search is assumed to be from min to ALL after
    - If no min frequency is provided, the search is assumed to be from ALL before to max
    - If both min and max are given, the search is a window from min to max
    """
    q = [curr_query]
    try:
        if params.startfreq and params.endfreq is None:
            # All images where max_frequency is greater than desired start
            q.append(f' AND max_frequency:[{params.startfreq:.9e} TO *]')
        elif params.endfreq and params.startfreq is None:
            # All images where min_frequency is less than desired end
            q.append(f' AND min_frequency:[* TO {params.endfreq:.9e}]')
        elif params.startfreq and params.endfreq:
            # All images:
            #  - completely within the desired window   (min_freq > start and max_freq < end)
            #  - straddling the windows bounds          (min_freq < start and max_freq > end)
            #  - stradding the desired start            (min_freq < start and max_freq < end and max_freq > start)
            #  - stradding the desired end              (min_freq > start and max_freq > end and min_freq < end)
            q.append(' AND ')
            q.append(f'((min_frequency:[{params.startfreq:.9e} TO *] AND max_frequency:[* TO {params.endfreq:.9e}])')
            q.append(' OR ')
            q.append(f'(min_frequency:[* TO {params.startfreq:.9e}] AND max_frequency:[{params.endfreq:.9e} TO *])')
            q.append(' OR ')
            q.append(f'(min_frequency:[* TO {params.startfreq:.9e}] AND max_frequency:[* TO {params.endfreq:.9e}]'
                     f' AND max_frequency:[{params.startfreq:.9e} TO *])')
            q.append(' OR ')
            q.append(f'(min_frequency:[{params.startfreq:.9e} TO *] AND max_frequency:[{params.endfreq:.9e} TO *]'
                     f' AND min_frequency:[* TO {params.endfreq:.9e}]))')
    except Exception as x:
        ex = 'Exception caught in {0}: {1}'.format(def_name(), str(x))
        raise Exception(ex)
    return ''.join(q)


def append_child_subquery(curr_query, parent, field, params,
                          append_operator='AND', field_join='AND', use_regex=False):
    """Append a search for a parameter under a child document.

    :param curr_query: a string carrying the current query string
    :param parent: the parent document of this child
    :param field: the child field to search against
    :param params: a list of parameters to search for
    :param append_operator: the operator to use in appending this subquery to the main query
    :param field_join: the operator to use in joining the parameters
    :param use_regex: if True, wrap value[x] with /.*"val".*/, otherwise it will be "val"
    :return: the query with the child query appended in the form of:
    ... [AND|OR] _query_:"{{!parent which=_docType_:<parent>}}
                                <field>:(\\"p1\\") [AND|OR] <field>:(\\"p2\\")..."
    or, if use_regex=True:
    ... [AND|OR] _query_:"{{!parent which=_docType_:<parent>}}
                                <field>:(/.*\\"p1\\".*/) [AND|OR] <field>:(/.*\\"p2\\".*/)..."
    """
    if append_operator not in ['AND', 'OR']:
        raise Exception('unknown append operator')
    if field_join not in ['AND', 'OR']:
        raise Exception('unknown join operator')
    q = [curr_query]
    try:
        if params:
            prefix = '/.*\\"' if use_regex else '\\"'
            suffix = '\\".*/' if use_regex else '\\"'
            search_list = ['{f}:({pre}{tok}{suf})'.format(f=field, pre=prefix, tok=x.strip(), suf=suffix)
                           for x in params.replace('\\',
                                                   '\\\\').replace('/', '\/').replace('"',
                                                                                      '\\"').replace(',', ' ').split()]
            if hasattr(search_list, '__len__') and isinstance(search_list, list) and len(search_list) > 0:
                q.append(' {0} _query_:"{{!parent which=_docType_:{1}}}'.format(append_operator, parent))
                q.append((' ' + field_join + ' ').join(search_list))
                q.append('"')
    except Exception as x:
        ex = 'Exception caught in {0}: {1}'.format(def_name(), str(x))
        raise Exception(ex)
    return ''.join(q)


def append_parent_subquery(curr_query, field, params,
                           append_operator='AND', field_join='AND', use_regex=False):
    """Append a search for a parameter under a child document.

    :param curr_query: a string carrying the current query string
    :param field: the child field to search against
    :param params: a list of parameters to search for
    :param append_operator: the operator to use in appending this subquery to the main query
    :param field_join: the operator to use in joining the parameters
    :param use_regex: if True, wrap value[x] with /.*"val".*/, otherwise it will be "val"
    :return: the query with the child query appended in the form of:
    ... [AND|OR] _query_:"{{!parent which=_docType_:<parent>}}
                                <field>:(\\"p1\\") [AND|OR] <field>:(\\"p2\\")..."
    or, if use_regex=True:
    ... [AND|OR] _query_:"{{!parent which=_docType_:<parent>}}
                                <field>:(/.*\\"p1\\".*/) [AND|OR] <field>:(/.*\\"p2\\".*/)..."
    """
    if append_operator not in ['AND', 'OR']:
        raise Exception('unknown append operator')
    if field_join not in ['AND', 'OR']:
        raise Exception('unknown join operator')
    q = [curr_query]
    try:
        if params:
            prefix = '/.*' if use_regex else ''
            suffix = '.*/' if use_regex else ''
            search_list = ['{f}:({pre}{tok}{suf})'.format(f=field, pre=prefix, tok=x.strip(), suf=suffix)
                           for x in params.replace('\\',
                                                   '\\\\').replace('/', '\/').replace('"',
                                                                                      '\\"').replace(',', ' ').split()]
            if hasattr(search_list, '__len__') and isinstance(search_list, list) and len(search_list) > 0:
                q.append(' {0} _query_:"{{!child of=_docType_:project}}'.format(append_operator))
                q.append((' ' + field_join + ' ').join(search_list))
                q.append('"')
    except Exception as x:
        ex = 'Exception caught in {0}: {1}'.format(def_name(), str(x))
        raise Exception(ex)
    return ''.join(q)


def append_polarization_subquery(curr_query, parent, field, params, append_operator='AND', field_join='AND',
                                 use_regex=False):
    """Append polarizations to the query.

    This is of the same form as the append_child_subquery above, but the polarizations are already in lists
    separated by commas, so, for the moment, we'll give them a little special treatment.
    :param curr_query: a string carrying the current query string
    :param parent: the parent document of this child
    :param field: the child field to search against
    :param params: a list of parameters to search for
    :param append_operator: the operator to use in appending this subquery to the main query
    :param field_join: the operator to use in joining the parameters
    :param use_regex: if True, wrap value[x] with /.*"val".*/, otherwise it will be "val"
    :return:
    """
    if append_operator not in ['AND', 'OR']:
        raise Exception('unknown append operator')
    if field_join not in ['AND', 'OR']:
        raise Exception('unknown join operator')
    q = [curr_query]
    try:
        if params:
            prefix = '/.*\\"' if use_regex else '\\"'
            suffix = '\\".*/' if use_regex else '\\"'
            search_list = ['{f}:({pre}{tok}{suf})'.format(f=field, pre=prefix, tok=x.strip(), suf=suffix)
                           for x in params]
            if hasattr(search_list, '__len__') and isinstance(search_list, list) and len(search_list) > 0:
                q.append(' {0} _query_:"{{!parent which=_docType_:{1}}}'.format(append_operator, parent))
                q.append((' ' + field_join + ' ').join(search_list))
                q.append('"')
    except Exception as x:
        ex = 'Exception caught in {0}: {1}'.format(def_name(), str(x))
        raise Exception(ex)
    return ''.join(q)


def append_field_search(curr_query, field_name, value, append_operator='AND', field_join='AND', use_regex=False):
    """ Append a search for words found in a string, with an OR or AND, depending on input.

    :param curr_query: The current query string we're going to append to and return.
    :param field_name: The solr field name we're querying.
    :param value: A string with zero or more words to search for in the field passed.
    :param append_operator: the operator to use when adding these parameters to the search; 'OR' or 'AND'
    :param field_join: the joining string "AND" or "OR" that we want to join the fields with.
    :param use_regex: if True, wrap value[x] with /.*"val".*/, otherwise it will be "val"
    :return: A query with the given string split into words and appended to the solr query in the form of:
    ... [AND|OR] field_name:(/.*word1.*/ [OR|AND] /.*word2.*/ [OR|AND] ...).
    or as
    ... [AND|OR] field_name:("word1" [OR|AND] "word2" [OR|AND] ...)
    """
    if append_operator not in ['AND', 'OR']:
        raise Exception('unknown append operator')
    if field_join not in ['AND', 'OR']:
        raise Exception('unknown join operator')
    q = [curr_query]
    field_list = build_field_list(value, field_join, use_regex)
    if field_list:
        q.append(' {0} {1}:({2})'.format(append_operator, field_name, field_list))
    return ''.join(q)


def build_field_list(values, field_join='AND', use_regex=False):
    """Build a list of search parameters with the field_join between them, optionally as a solr regex search

    Join the parameters in the values list into a formatted list separated by the field_join, as:
    "val1" field_join "val2" field_join "val3"...
    or, if use_regex=True:
    /.*"val1".*/ field_join /.*"val2".*/ field_join /.*"val3".*/...

    :param values: a string of parameters to search for, separated by whitespace
    :param field_join: the logical operator to use in joining the search parameters
    :param use_regex: if True, wrap value[x] with /.*"val".*/, otherwise it will be "val"
    :return: return the formatted list of parameters
    """
    if values:
        prefix = '/.*"' if use_regex else '"'
        suffix = '".*/' if use_regex else '"'
        search_list = ['{pre}{tok}{suf}'.format(pre=prefix, tok=x.replace('\\', '\\\\').replace('/', '\/').replace('"',
                                                                                                                   '\\"').strip(),
                                                suf=suffix)
                       for x in
                       values.replace(',', ' ').split()]
        if hasattr(search_list, '__len__') and isinstance(search_list, list) and len(search_list) > 0:
            return (' ' + field_join + ' ').join(search_list)
    return None


def build_paired_field_list(values, field_join='AND'):
    """Build a list of strings to search for, as a pair of ("text" OR /.*text.*/); tokenized string OR regex search

    Because solr will apply the tokenizers to strings in searches, but not regex expressions, we use both forms
    to maximize the possibility of getting the user a result.
    :param values: a string of parameters to search for, separated by whitespace
    :param field_join: the logical operator to use in joining the search parameters
    :return: return the formatted list of parameters
    """
    if values:
        prefix = '/.*"'
        suffix = '".*/'
        search_list = ['("{tok}" OR {pre}{tok}{suf})'.format(pre=prefix, tok=re.sub(r'[\W]', '', x.strip()), suf=suffix)
                       for x in values.replace(',', ' ').split()]
        if hasattr(search_list, '__len__') and isinstance(search_list, list) and len(search_list) > 0:
            return (' ' + field_join + ' ').join(search_list)
    return None


def append_textblob_search(curr_query, field, text_search_input, field_join='AND'):
    """Append a query for a search against the _text field of the document.

    We are using the _text field to provide a google-esque search for keywords.  To achieve this we use the
    power of solr's tokenized search and their regex searches.  We

    :param curr_query: a string carrying the current query string
    :param text_search_input: the string of words the user is searching for
    :return: the query with the text_search_input words appended in the form:
    ... [AND|OR] (_text:(("word1" OR /.*"word1".*/) AND ("word2" OR /.*"word3".*/) AND ... ("wordn" OR /.*"wordn".*/)))
    """
    q = [curr_query]
    if text_search_input:
        append_operator = 'AND'

        q.append(
            ' {0} ({1}:({2})'.format(append_operator, field, build_paired_field_list(text_search_input, field_join)))
        q.append(')')

    return ''.join(q)


def append_proj_column_search(curr_query, field_name, value):
    """ Append a search for words found in a text string, meant for front end column filters.

    :param curr_query: The current query string we're going to append to and return.
    :param field_name: The solr field name we're querying.
    :param value: A string with zero or more words to search for in the field passed.
    :return: A query with the given string split into words and appended to the solr query in the form of:
    ... AND field_name:(word1.*/ AND word2.*/ AND ...)
    This is a minor variant of "append_field_search", where instead of prefixing the words with wildcards, the
    words only wildcard the 'remaining' words.  This probably doesn't need to be adding lists, since this is meant for
    column filters on the front end, and should never see more than one word (but should is the key word, so leave it).
    """
    q = [curr_query]
    if value:
        prefix = '/"'
        suffix = '".*/'
        search_list = ['{pre}{tok}{suf}'.format(pre=prefix, tok=x.strip(), suf=suffix)
                       for x in
                       value.replace('\\', '\\\\').replace('/', '\/').replace('"', '\\"').replace(',', ' ').split()]
        if hasattr(search_list, '__len__') and isinstance(search_list, list) and len(search_list) > 0:
            q.append(' AND {0}:('.format(field_name))
            q.append(' AND '.join(search_list))
            q.append(')')
    return ''.join(q)


def append_regex_projects_codes(curr_query, values):
    """ Append a search for project codes, passed as a string or a list, to the current query.

    :param curr_query: The current query string we're going to append to and return.
    :param values: A string or list with zero or more project codes in it.
    :return: A query with the project codes split and added in list form to the solr query.  This is different than
    "append_text_search_to_query", because 1) we're silently added a cross-reference search against the legacy_id (So,
    for each project_code passed, we assume the user could have meant to search for a legacy VLA project, using a AB123
    format, or newer project codes like 17A-123), and 2) we are ORing the list of search "words".
    """
    if not values:
        return curr_query
    else:
        q = [curr_query]
        if hasattr(values, '__len__') and (isinstance(values, list)):
            escaped_list = [
                '/.*"{0}".*/'.format(v.replace('\\', '\\\\').replace('/', '\/').replace('"', '\\"').replace(',', ' '))
                for v in values]
            q.append(' AND (')
            if len(values) > 1:
                ored_string = ' OR '.join(escaped_list)
                combined_str = ''.join(['project_code:(', ored_string, ') OR legacy_id:(', ored_string, ')'])
            else:
                combined_str = 'project_code:{0} OR legacy_id:{0}'.format(''.join(escaped_list[0]))
            q.append(combined_str)
            q.append(')')
        else:
            q.append(' AND (project_code:/.*"{0}".*/ OR legacy_id:/.*"{0}".*/)'.format(
                values.replace('\\', '\\\\').replace('/', '\/').replace('"', '\\"').replace(',', ' ')))
    return ''.join(q)


def construct_execution_block_query(params):
    """Build a query object with a query string and a geofilt object for position.

    Using the parameters passed in, build a query against exec_block documents in the solr.execblocks core.
    :param params: a dictionary of parameters to use in constructing the search query
    :return: an object with a query string and a position filter object
    """
    q = '_docType_:exec_block'
    geofilt = None
    if params is not None:

        try:
            q = append_field_search(q, 'instrument_name', ' '.join(params.instrument_name), 'AND', 'OR', True)
            q = append_field_search(q, 'vla_configuration', ' '.join(params.vla_configuration), 'AND', 'OR', False)
            q = append_field_search(q, 'obs_band', ' '.join(params.obs_band), 'AND', 'OR', False)
            q = append_field_search(q, 'pi', params.pi, 'AND', 'OR', True)
            q = append_field_search(q, 'title', params.title, 'AND', 'OR', False)
            q = append_field_search(q, 'abstract', params.abstract, 'AND', 'OR', False)
            q = append_field_search(q, 'obs_id', ' '.join(params.obs_id), 'AND', 'OR', True)

            q = append_field_search(q, 'mousId', params.mousId, 'AND', 'OR', False)

            q = append_child_subquery(q, 'exec_block', 'target_name', params.target_name, 'AND', 'OR', True)
            q = append_polarization_subquery(q, 'exec_block', 'polarizations', params.polarizations, 'AND', 'OR', False)
            q = append_date_subquery(q, params)
            q = append_eb_frequency_subquery(q, params)
            q = append_textblob_search(q, '_text', params.text_search_str, 'AND')
            q = append_regex_projects_codes(q, params.project_code)
            q = get_alternate_eb_position_filter(q, params)
            # geofilt = get_eb_position_filter(params)

            if params.show_cms_only:
                q = append_field_search(q, 'cal_status', 'Calibrated', 'AND', 'OR')

            if params.show_public_only:
                q = append_field_search(q, 'data_rights', 'PUBLIC', 'AND', 'OR')

            # log.info(f'q: {q}')
        except Exception as x:
            raise x
    return dict(q=q, geofilt=geofilt)


def construct_images_query(params):
    """Build a query object with a query string and a geofilt object for position.

    Using the parameters passed in, build a query against project documents in the solr.images core.
    :param params: a dictionary of parameters to use in constructing the search query
    :return: an object with a query string and a position filter object
    """
    q = '_docType_:image'
    geofilt = None
    if params is not None:
        try:
            # There isn't a search filter on the front end for this, but it's here when I figure out how to
            # wire in view specific filters.
            q = append_field_search(q, 'instrument_name', ' '.join(params.instrument_name), 'AND', 'OR', True)
            q = append_field_search(q, 'band_code', ' '.join(params.obs_band), 'AND', 'OR', False)

            q = append_field_search(q, 'target_name', params.target_name, 'AND', 'OR', False)
            q = append_field_search(q, 'polarization', ' '.join(params.polarizations), 'AND', 'OR', True)

            q = append_field_search(q, 'project_code', ' '.join(params.project_code), 'AND', 'OR', True)
            q = append_field_search(q, 'file_name', ' '.join(params.obs_id), 'AND', 'OR', True)
            q = append_images_frequency_subquery(q, params)
            geofilt = get_images_position_filter(params)

            # log.info(f'q: {q}')
        except Exception as x:
            raise x

    return dict(q=q, geofilt=geofilt)


def append_eb_column_filters_query(q, params):
    """Append additional search parameters based on a column_filters parameter.

    This couples the solr_interface to a design feature in the archive front end, where
    the execblocks core is presented to the user grouped by projects and tables
    listing exec_blocks under the project.  This allows for filtering those tables
    based on a column filter accepting partial strings.
    :param q: a string carrying the current query string
    :param params: the parameters passed in for constructing a query
    :return: the query with the column filter parameters appended, with trailing regex wildcards.
    """
    if params is not None:
        if 'obs_id' in params.column_filters:
            q = append_proj_column_search(q, 'obs_id', params.column_filters['obs_id'])
        if 'cal_status' in params.column_filters:
            q = append_proj_column_search(q, 'cal_status', params.column_filters['cal_status'])

    return q


class ArchiveSolrParameters(object):
    """
        This is a container object for holding the solr query parameters.
    """

    def __repr__(self):
        """ Return the string representation of this class.

        :return: The string representation of self.
        """
        repr_str_list = ['{c}(']
        temp = ['{0}={{s.{0}}}'.format(key) for key in vars(self)]
        self_list = ', '.join(temp)
        repr_str_list.append(self_list)
        repr_str_list.append(')')

        return ''.join(repr_str_list).format(c=ArchiveSolrParameters.__name__, s=self)

    def __init__(self, **kwargs):

        self.start = kwargs.get('start', 0)  # start if start else 0
        self.rows = kwargs.get('rows', None)
        self.sort = kwargs.get('sort', None)

        self.mindate = kwargs.get('mindate', None)
        self.maxdate = kwargs.get('maxdate', None)

        self.solr_id = kwargs.get('solr_id', None)
        self.target_name = kwargs.get('target_name', None)
        self.pi = kwargs.get('pi', None)
        self.title = kwargs.get('title', None)
        self.abstract = kwargs.get('abstract', None)

        self.text_search_str = kwargs.get('text_search_str', None)

        # For now, we only expect these four to come in as comma separated lists, with polarization a special case
        # of comma-space-containing members in the list, which need to retain that comma-space delimiting characteristic
        # or the solr search will miss the combinations.  So, rather than
        self.instrument_name = re.split(r',(?! )', json.loads(kwargs.get('instrument_name'))) \
            if kwargs.get('instrument_name', None) else []
        try:
            # try to address the index 'VLA', and if that throws a key error, it's not in params, move on
            # otherwise replace it with 'EVLA'
            self.instrument_name.index('VLA')
            self.instrument_name.append('EVLA')
        except ValueError:
            pass

        self.vla_configuration = re.split(r',(?! )', json.loads(kwargs.get('vla_configuration'))) \
            if kwargs.get('vla_configuration', None) else []
        self.obs_band = re.split(r',(?! )', json.loads(kwargs.get('obs_band'))) \
            if kwargs.get('obs_band', None) else []
        self.polarizations = re.split(r',(?! )', json.loads(kwargs.get('polarizations'))) \
            if kwargs.get('polarizations', None) else []

        # We're treating this as though we have a multi-select pulldown for this like instrument_name, but we
        # don't have any control implemented for this yet.
        self.collection_names = re.split(r',(?! )', json.loads(kwargs.get('collection_names'))) \
            if kwargs.get('collection_names', None) else []

        self.project_code = [p for p in re.split(r'\s*[ |,]\s*', json.loads(kwargs.get('project_code', None))) if p and len(p) > 0] \
                if kwargs.get('project_code', None) else []
        self.obs_id = [o for o in re.split(r'\s*[ |,]\s*', json.loads(kwargs.get('obs_id', None))) if o and len(o) > 0] \
            if kwargs.get('obs_id', None) else []

        self.mousId = kwargs.get('mousId', None)

        self.coordsys_refFrame = kwargs.get('coordsys_refFrame').lower() if kwargs.get('coordsys_refFrame',
                                                                                       None) else None
        self.radius = float(kwargs.get('radius')) if kwargs.get('radius', None) else None
        self.equinox = kwargs.get('equinox', 'J2000')
        RA = float(kwargs.get('longitude')) if kwargs.get('longitude', None) else None
        Dec = float(kwargs.get('latitude')) if kwargs.get('latitude', None) else None

        if (RA is not None and Dec is None) or (RA is None and Dec is not None):
            ex = 'Exception caught in {0}: {1}'.format(def_name(), 'position coordinates must be given in pairs')
            raise Exception(ex)
        elif RA is None and Dec is None:
            self.longitude = None
            self.latitude = None
        else:
            if self.equinox != 'J2000':
                B1950_coords = SkyCoord(ra=RA * u.deg, dec=Dec * u.deg, frame=FK4, equinox='B1950')
                J2000_coords = B1950_coords.transform_to(ICRS)
                self.longitude = J2000_coords.ra.deg - 180
                self.latitude = J2000_coords.dec.deg
            else:
                self.longitude = RA - 180
                self.latitude = Dec

        self.startfreq = float(kwargs.get('startfreq')) if kwargs.get('startfreq', None) else None
        self.endfreq = float(kwargs.get('endfreq')) if kwargs.get('endfreq', None) else None

        self.column_filters = json.loads(kwargs.get('column_filters')) if kwargs.get('column_filters', None) else None

        self.show_cms_only = True if kwargs.get('show_cms_only', None) == 'true' else False
        self.show_public_only = True if kwargs.get('show_public_only', None) == 'true' else False
